/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <unistd.h>
#include <time.h>
#include <limits.h>
#include <fcntl.h>
#include <dirent.h>
#include <vconf.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <vconf.h>
#include <fcntl.h>
#include <sys/syscall.h>
#include <linux/reboot.h>
#include <sys/time.h>
#include <mntent.h>
#include <sys/mount.h>
#include <device-node.h>
#include <bundle.h>
#include <eventsystem.h>
#include <libgdbus/dbus-system.h>

#include "dd-deviced.h"
#include "core/log.h"
#include "core/launch.h"
#include "core/config-parser.h"
#include "core/device-notifier.h"
#include "core/device-idler.h"
#include "core/common.h"
#include "core/devices.h"
#include "core/launch.h"
#include "display/poll.h"
#include "display/setting.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "power-handler.h"
#include "apps/apps.h"
#include "shared/deviced-systemd.h"
#include "boot.h"

#define POWEROFF_DURATION		2
#define MAX_RETRY			2
#define POWEROFF_WAIT_RESOURCED		(0.5*1000) /* 0.5 seconds */
#define POWEROFF_WAIT_MAX		10 /* 10 seconds */

#define SIGNAL_POWEROFF_STATE		"ChangeState"

#define UMOUNT_RW_PATH_USER		"/opt/usr"
#define UMOUNT_RW_PATH_SYSTEM		"/opt"
#define MAX_UMOUNT_KILL_RETRY		4

#define POWER_FLAG_POWEROFF		"/run/"POWER_POWEROFF
#define POWER_FLAG_REBOOT		"/run/"POWER_REBOOT

#define POWER_CONF_FILE			"/etc/deviced/power.conf"

struct power_flag {
	enum poweroff_type type;
	const char *path;
} power_flags[] = {
	{ POWER_OFF_DIRECT,  POWER_FLAG_POWEROFF },
	{ POWER_OFF_RESTART, POWER_FLAG_REBOOT   },
};

static struct timeval tv_start_poweroff;

static dd_list *poweroff_options;
static struct power_option poweroff_opt;
static bool during_poweroff;
static dd_list *poweroff_handles;

static int get_poweroff_type_enum(char *type, enum poweroff_type *val)
{
	size_t len;

	if (!type || !val)
		return -EINVAL;

	len = strlen(type) + 1;

	if (!strncmp(POWER_POWEROFF, type, len)) {
		*val = POWER_OFF_DIRECT;
		return 0;
	}

	if (!strncmp(PWROFF_POPUP, type, len)) {
		*val = POWER_OFF_POPUP;
		return 0;
	}

	if (!strncmp(POWER_REBOOT, type, len)) {
		*val = POWER_OFF_RESTART;
		return 0;
	}

	_E("Invalid poweroff type (%s)", type);
	return -EINVAL;
}

static int find_poweroff_option(char *type, char *option)
{
	int ret;
	dd_list *l;
	struct power_option *elem;
	enum poweroff_type type_e;

	if (!type)
		return -EINVAL;

	ret = get_poweroff_type_enum(type, &type_e);
	if (ret < 0) {
		_E("Failed to get type enum value (%d)", ret);
		return ret;
	}

	DD_LIST_FOREACH(poweroff_options, l, elem) {
		if (elem->type != type_e)
			continue;
		if (option) {
			if (elem->option == NULL)
				continue;
			if (strncmp(elem->option, option, strlen(elem->option)))
				continue;
		} else {
			if (elem->option != NULL)
				continue;
		}
		poweroff_opt.type = elem->type;
		if (poweroff_opt.name) {
			free(poweroff_opt.name);
			poweroff_opt.name = NULL;
		}
		poweroff_opt.name = (elem->name ? strdup(elem->name) : NULL);
		if (poweroff_opt.option) {
			free(poweroff_opt.option);
			poweroff_opt.option = NULL;
		}
		poweroff_opt.option = (option ? strdup(option) : NULL);
		return 0;
	}

	return -ENOENT;
}

static int stop_systemd_journald(void)
{
	int ret;

	ret = deviced_systemd_stop_unit("systemd-journald.socket");
	if (ret < 0) {
		_E("failed to stop 'systemd-journald.socket'");
		return ret;
	}

	ret |= deviced_systemd_stop_unit("systemd-journald.service");
	if (ret < 0) {
		_E("failed to stop 'systemd-journald.service'");
		return ret;
	}
	_I("done %s with %s option",
		poweroff_opt.name ? poweroff_opt.name : "NULL",
		poweroff_opt.option ? poweroff_opt.option : "NULL");
	return 0;
}

/* umount usr data partition */
static void unmount_rw_partition(void)
{
	umount_partition_by_kill(UMOUNT_RW_PATH_USER, MAX_UMOUNT_KILL_RETRY);
	umount_partition_by_kill(UMOUNT_RW_PATH_SYSTEM, MAX_UMOUNT_KILL_RETRY);
}

static void powerdown(void)
{
	static int wait;
	struct timeval now;
	int poweroff_duration = POWEROFF_DURATION;
	int check_duration = 0;
	char *buf;

	stop_systemd_journald();
	sync();

	buf = getenv("PWROFF_DUR");
	if (buf != NULL && strlen(buf) < 1024)
		poweroff_duration = atoi(buf);
	if (poweroff_duration < 0 || poweroff_duration > 60)
		poweroff_duration = POWEROFF_DURATION;
	gettimeofday(&now, NULL);
	check_duration = now.tv_sec - tv_start_poweroff.tv_sec;
	while (check_duration < poweroff_duration) {
		if (wait == 0) {
			_I("wait poweroff %d %d", check_duration, poweroff_duration);
			wait = 1;
		}
		usleep(100000);
		gettimeofday(&now, NULL);
		check_duration = now.tv_sec - tv_start_poweroff.tv_sec;
		if (check_duration < 0)
			break;
	}

	watchdog_notify();

	if (!is_emulator())
		unmount_rw_partition();
}

static void reboot_with_option(int cmd, char *option)
{
	if (!cmd)
		return;
	if (syscall(SYS_reboot, LINUX_REBOOT_MAGIC1, LINUX_REBOOT_MAGIC2,
				cmd, (void *)option) < 0)
		_E("Failed to reboot with option (%d)", errno);
}

static void poweroff(void)
{
	_I("Power off !!!");
	reboot_with_option(LINUX_REBOOT_CMD_POWER_OFF, NULL);
}

static void restart(char *option)
{
	if (option) {
		_I("Restart with option %s !!!", option);
		reboot_with_option(LINUX_REBOOT_CMD_RESTART2, option);
	} else {
		_I("Restart !!!");
		reboot_with_option(LINUX_REBOOT_CMD_RESTART, NULL);
	}
}

void powerdown_ap(void)
{
	_I("%s option(%s)",
		poweroff_opt.name ? poweroff_opt.name : "NULL",
		poweroff_opt.option ? poweroff_opt.option : "NULL");
	powerdown();

	if (poweroff_opt.type == POWER_OFF_DIRECT) {
		poweroff();
		return;
	}

	if (poweroff_opt.type == POWER_OFF_RESTART) {
		restart(poweroff_opt.option);
		return ;
	}

	_E("poweroff invalid type (%d)", poweroff_opt.type);
}

static void system_shutdown_send_system_event(void)
{
	bundle *b;
	const char *str = EVT_VAL_SYSTEM_SHUTDOWN_TRUE;

	b = bundle_create();
	bundle_add_str(b, EVT_KEY_SYSTEM_SHUTDOWN, str);
	eventsystem_send_system_event(SYS_EVENT_SYSTEM_SHUTDOWN, b);
	bundle_free(b);
}

static int pwroff_popup(void)
{
	return launch_system_app(APP_POWERKEY,
			2, APP_KEY_TYPE, APP_POWERKEY);
}

static void disable_display(void)
{
	const struct device_ops *display_device_ops = NULL;
	FIND_DEVICE_VOID(display_device_ops, "display");
	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_POWEROFF, LCD_NORMAL);
	display_device_ops->exit(NULL);
}

static int poweroff_procedure(void)
{
	disable_display();
	watchdog_notify();
	powerdown_ap();
	return 0;
}

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		goto out;

	done = *(int *)data;
out:
	return done;
}

static void poweroff_send_broadcast(int status)
{
	static int old = 0;

	if (old == status)
		return;

	_D("broadcast poweroff %d", status);

	old = status;

	/* Need to notify to deviced-vibrator. deviced-vibrator receives ChangeState signal for POWER_OFF_DIRECT and POWER_OFF_RESTART */
	dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_POWEROFF,
						DEVICED_INTERFACE_POWEROFF,
						SIGNAL_POWEROFF_STATE,
						g_variant_new("(i)", status));
}

int check_power_flag(void)
{
	char buf[32] = { 0, };
	char *option = NULL;
	int i, index;
	int fd;
	ssize_t len;

	index = -1;
	for (i = 0 ; i < ARRAY_SIZE(power_flags) ; i++) {
		if (access(power_flags[i].path, F_OK) == 0) {
			index = i;
			break;
		}
	}
	if (index < 0)
		return 0;

	if (power_flags[index].type == POWER_OFF_DIRECT) {
		poweroff();
		return 1;
	}

	fd = open(power_flags[index].path, O_RDONLY);
	if (fd >= 0) {
		len = read(fd, buf, sizeof(buf) - 1);
		close(fd);
		if (len < 0) {
			len = errno;
			_E("read() failed (%d)", len);
			return 0;
		}
		while (len > 0 && buf[len - 1] == '\n')
			len--;
		buf[len] = '\0';
		_I("reboot option (%s)", buf);
	}

	if (strlen(buf) > 0)
		option = buf;

	restart(option);

	return 1;
}

static void make_power_flag(enum poweroff_type type, char *option)
{
	const char *path;
	int fd;
	ssize_t len;
	int i;

	path = NULL;
	for (i = 0 ; i < ARRAY_SIZE(power_flags) ; i++) {
		if (power_flags[i].type == type) {
			path = power_flags[i].path;
			break;
		}
	}
	if (!path)
		return;

	fd = open(path, O_RDWR|O_CREAT, S_IRUSR | S_IWUSR);
	if (fd < 0) {
		_E("Failed to create (%s)", path);
		return;
	}

	if (option) {
		len = write(fd, option, strlen(option));
		if (len <= 0)
			_E("Failed to store option (%d)", len < 0 ? errno : len);
	}

	close(fd);
}

static void poweroff_start_animation(void)
{
	int ret;

	ret = deviced_systemd_start_unit("shutdown-animation.service");
	if (ret < 0)
		_E("Failed to start shutdown animation");

	gettimeofday(&tv_start_poweroff, NULL);
}

static void poweroff_notify_resourced(void)
{
	_I("Request to stop systemd service to resourced");
	dbus_handle_method_sync_timeout(RESOURCED_BUS_NAME,
				RESOURCED_PATH_PROCESS,
				RESOURCED_INTERFACE_PROCESS,
				"PrePoweroff",
				NULL,
				NULL,
				POWEROFF_WAIT_RESOURCED);
}

/*
 * Notice : Don't change sequence of poweroff_fixed_flow
 *          It will make blocking issue during poweroff
 */
static void poweroff_fixed_flow(enum poweroff_type type)
{
	int off = (int)type;

	poweroff_start_animation();
	poweroff_notify_resourced();
	device_notify(DEVICE_NOTIFIER_POWEROFF, &off);
}

static void poweroff_remove_handle(pid_t pid)
{
	struct poweroff_handle *handle;
	dd_list *elem;

	DD_LIST_FOREACH(poweroff_handles, elem, handle) {
		if (handle->pid == pid)
			break;
	}

	assert(handle);

	_D("Remove handle pid(%d) timeout(%d) timeout_id(%d)", handle->pid, handle->timeout, handle->timeout_id);
	DD_LIST_REMOVE(poweroff_handles, handle);

	if (handle->timeout_id) {
		g_source_remove(handle->timeout_id);
		handle->timeout = 0;
		handle->timeout_id = 0;
	}

	free(handle);
}

static gboolean poweroff_timeout_cb(void *data)
{
	int ret;
	pid_t pid;

	pid = (pid_t)((intptr_t)data);
	poweroff_remove_handle(pid);

	if (DD_LIST_LENGTH(poweroff_handles)) {
		_D("Timer is left");
		return G_SOURCE_REMOVE;
	}
	_D("No timer left");

	if (!poweroff_opt.name) {
		_E("There is no option");
		goto out;
	}

	watchdog_notify();

	if (poweroff_opt.type != POWER_OFF_DIRECT &&
		poweroff_opt.type != POWER_OFF_RESTART) {
		_E("Invalid type (%d)", poweroff_opt.type);
		goto out;
	}

	during_poweroff = true;

	make_power_flag(poweroff_opt.type, poweroff_opt.option);

	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_POWEROFF, LCD_OFF, STAY_CUR_STATE, 0);

	poweroff_fixed_flow(poweroff_opt.type);

	/* TODO for notify. will be removed asap. */
	vconf_set_int(VCONFKEY_SYSMAN_POWER_OFF_STATUS, poweroff_opt.type);

	ret = poweroff_procedure();
	if (ret < 0)
		_E("Failed to %s (%d)", poweroff_opt.name, ret);

out:
	if (disp_plgn.update_pm_setting)
		disp_plgn.update_pm_setting(SETTING_POWEROFF, poweroff_opt.type);
	during_poweroff = false;
	return G_SOURCE_REMOVE;
}

void poweroff_start_timers()
{
	struct poweroff_handle *handle = NULL;
	dd_list *l;
	bool timer_exist = false;
	int pid_alive = 0;

	DD_LIST_FOREACH(poweroff_handles, l, handle) {
		pid_alive = kill(handle->pid, 0);
		if (pid_alive == -1) {
			_D("Pid(%d) is dead", handle->pid);
			handle->timeout = 0;
		}

		_D("Run timer, pid(%d) timeout(%d) timeout_id(%d)", handle->pid, handle->timeout, handle->timeout_id);
		handle->timeout_id = g_timeout_add_seconds(handle->timeout,
				poweroff_timeout_cb,
				(void *)((intptr_t)(handle->pid)));

		timer_exist = true;
	}

	if (timer_exist) {
		return;
	} else {
		_D("handle is NULL");
		handle = (struct poweroff_handle *)malloc(sizeof(struct poweroff_handle));
		if (handle == NULL) {
			_E("Not enough memory.");
			return;
		}

		handle->pid = getpid();
		handle->timeout = 0;
		handle->timeout_id = g_timeout_add_seconds(handle->timeout,
				poweroff_timeout_cb,
				(void *)((intptr_t)(handle->pid)));

		if (!handle->timeout_id) {
			_E("timer_add failed");
			free(handle);
			return;
		}

		DD_LIST_APPEND(poweroff_handles, handle);
	}

	_D("Last Timer : timer_id(%d) pid(%d) timeout(%d)", handle->timeout_id, handle->pid, handle->timeout);
	return;
}

static int power_execute_pid(char *type, char *option, pid_t pid)
{
	int ret;

	if (during_poweroff) {
		_E("during poweroff");
		return -EINVAL;
	}

	if (!type)
		return -EINVAL;

	ret = find_poweroff_option(type, option);
	if (ret < 0) {
		_E("Failed to find supported %s option %s (%d)", type, (option ? option : "NULL"), ret);
		return ret;
	}

	if (strncmp(PWROFF_POPUP, poweroff_opt.name, PWROFF_POPUP_LEN) == 0) {
		_D("request popup");
		if (poweroff_opt.type == POWER_OFF_POPUP) {
			vconf_set_int(VCONFKEY_SYSMAN_POWER_OFF_STATUS, poweroff_opt.type);
			pwroff_popup();
		}
		return 0;
	}

	poweroff_opt.pid = pid;
	/* Poweroff event broadcasting */
	system_shutdown_send_system_event();
	poweroff_send_broadcast(poweroff_opt.type);
	poweroff_start_timers();

	return 0;
}

static int power_execute(void *data)
{
	return power_execute_pid((char *)data, NULL, getpid());
}

GVariant *dbus_power_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	pid_t pid;
	int ret;
	int argc;
	char *type_str;

	g_variant_get(param, "(si)", &type_str, &argc);

	if (argc < 0) {
		_E("message is invalid!");
		ret = -EINVAL;
		goto out;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("%d process does not exist, dbus ignored!", pid);
		ret = -ESRCH;
		goto out;
	}

	_I("poweroff PID(%d) requests %s", pid, type_str);
	ret = power_execute_pid(type_str, NULL, pid);

out:
	g_free(type_str);
	return g_variant_new("(i)", ret);
}

GVariant *dbus_power_option_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	pid_t pid;
	int ret;
	char *type, *option;

	g_variant_get(param, "(ss)", &type, &option);

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("%d process does not exist, dbus ignored!", pid);
		ret = -ESRCH;
		goto out;
	}

	_I("poweroff PID(%d) requests %s %s", pid, type, option);
	ret = power_execute_pid(type, option, pid);

out:
	g_free(type);
	g_free(option);
	return g_variant_new("(i)", ret);
}

GVariant *request_reboot(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	char *str;
	int ret;
	pid_t pid;

	g_variant_get(param, "(s)", &str);

	pid = dbus_connection_get_sender_pid(conn, sender);

	_I("poweroff PID(%d) requests reboot with command : %s", pid, str);
	ret = power_execute_pid(POWER_REBOOT, NULL, pid);

	g_free(str);
	return g_variant_new("(i)", ret);
}

static GVariant *add_poweroff_time(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct poweroff_handle *handle;
	int ret;
	pid_t pid;
	dd_list *l;

	pid = dbus_connection_get_sender_pid(conn, sender);
	_D("add_poweroff_timer(%d)", pid);
	if (pid == -1) {
		_E("Failed to get sender's pid.");
		ret = -1;
		goto add_out;
	}

	DD_LIST_FOREACH(poweroff_handles, l, handle) {
		if (handle->pid == pid)
			break;
	}

	if (handle)
		poweroff_remove_handle(pid);

	_D("Make a new handle");
	handle = (struct poweroff_handle *)malloc(sizeof(struct poweroff_handle));
	if (handle == NULL) {
		_E("Not enough memory.");
		ret = -1;
		goto add_out;
	}

	handle->pid = pid;
	handle->timeout_id = 0;
	handle->timeout = POWEROFF_WAIT_MAX;

	DD_LIST_APPEND(poweroff_handles, handle);
	_D("Add a new poweroff timer(%d %d %d)", handle->pid, handle->timeout, handle->timeout_id);

	ret = 0;

add_out:
	return g_variant_new("(i)", ret);
}

static GVariant *remove_poweroff_time(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	struct poweroff_handle *handle;
	dd_list *l;
	int ret = 0;
	pid_t pid;

	pid = dbus_connection_get_sender_pid(conn, sender);
	_D("remove_poweroff_timer(%d)", pid);
	if (pid == -1) {
		_E("Failed to get sender's pid.");
		ret = -1;
		goto remove_out;
	}

	DD_LIST_FOREACH(poweroff_handles, l, handle) {
		if (handle->pid == pid) {
			break;
		}
	}

	if (handle) {
		assert(handle);

		if (handle && handle->timeout_id) {
			g_source_remove(handle->timeout_id);
			handle->timeout = 0;
			handle->timeout_id = g_timeout_add_seconds(handle->timeout,
					poweroff_timeout_cb,
					(void *)((intptr_t)(handle->pid)));
		}
	} else {
		_E("Invalid pid(%d)", pid);
		ret = -1;
	}

remove_out:
	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ POWER_REBOOT, "si", "i", dbus_power_handler },
	{ "RebootWithOption", "ss", "i", dbus_power_option_handler },
	/* Public API device_power_reboot() calls this dbus method. */
	{ "Reboot",      "s", "i", request_reboot },
	{ "AddPoweroffWait",  NULL, "i", add_poweroff_time },
	{ "RemovePoweroffWait",  NULL, "i", remove_poweroff_time },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_REBOOT,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int add_poweroff_option(enum poweroff_type type, char *option)
{
	struct power_option *opt;
	char *name;

	if (type == POWER_OFF_POPUP)
		name = PWROFF_POPUP;
	else if (type == POWER_OFF_DIRECT)
		name = POWER_POWEROFF;
	else if (type == POWER_OFF_RESTART)
		name = POWER_REBOOT;
	else {
		_E("Invalid type (%d)", type);
		return -EINVAL;
	}

	opt = calloc(1, sizeof(struct power_option));
	if (!opt) {
		_E("calloc() failed");
		return -ENOMEM;
	}

	opt->type = type;
	opt->name = strdup(name);
	opt->option = (option ? strdup(option) : NULL);

	DD_LIST_APPEND(poweroff_options, opt);

	_D("Add %s option (%s)", opt->name, opt->option);

	return 0;
}

static int load_config(struct parse_result *result, void *user_data)
{
	enum poweroff_type type;
	int ret;

	if (MATCH(result->section, "Poweroff"))
		type = POWER_OFF_DIRECT;
	else if (MATCH(result->section, "Reboot"))
		type = POWER_OFF_RESTART;
	else
		return 0;

	if (!MATCH(result->name, "Option"))
		return 0;

	ret = add_poweroff_option(type, result->value);
	if (ret < 0) {
		_E("Failed to add %s option (%s)", result->section, result->value);
		return ret;
	}

	return 0;
}

static void power_init(void *data)
{
	int ret;

	/* init dbus interface */
	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_REBOOT, &dbus_interface);
	if (ret < 0)
		_E("fail to init dbus method(%d)", ret);

	add_booting_done_handler(NULL);

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);

	add_poweroff_option(POWER_OFF_DIRECT, NULL);
	add_poweroff_option(POWER_OFF_RESTART, NULL);
	add_poweroff_option(POWER_OFF_POPUP, NULL);

	ret = config_parse(POWER_CONF_FILE, load_config, NULL);
	if (ret < 0)
		_E("Failed to load power off config (%d)", ret);
}

static const struct device_ops power_device_ops = {
	.name     = POWER_OPS_NAME,
	.init     = power_init,
	.execute  = power_execute,
};

DEVICE_OPS_REGISTER(&power_device_ops)
