/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <vconf.h>
#include <bundle.h>
#include <eventsystem.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "display/poll.h"
#include "display/display-ops.h"
#include "extcon/extcon.h"
#include "apps/apps.h"
#include "usb.h"
#include "usb-tethering.h"
#include "usb-debug.h"

static struct usb_gadget_translator *gadget_translator;
static struct usb_client *usb_client;

struct extcon_ops extcon_usb_ops;

static struct usb_gadget_translator *gadget_translator_probe(void)
{
	struct hw_info *info;
	struct hw_common *common;
	struct usb_gadget_translator *gadget;

	if (hw_get_info(USB_GADGET_DEVICE_ID, (const struct hw_info **)&info)) {
		_E("No usb gadget translator");
		return NULL;
	}

	if (!info->open) {
		_E("Failed to open usb gadget translator; open(NULL)");
		return NULL;
	}

	if (info->open(info, NULL, &common) < 0) {
		_E("Failed to open usb gadget translator");
		return NULL;
	}

	gadget = container_of(common, struct usb_gadget_translator, common);
	if (!gadget->id_to_gadget || !gadget->cleanup_gadget) {
		_E("Invalid usb gadget translator");

		if (gadget->common.info)
			gadget->common.info->close(&gadget->common);

		return NULL;
	}

	return gadget;
}

static struct usb_client *usb_client_probe(const char *id)
{
	struct hw_info *info;
	struct hw_common *common;
	struct usb_client *client;

	_I("Loading usb client: %s", id);

	if (hw_get_info(id, (const struct hw_info **)&info)) {
		_I("No usb_client: %s", id);
		return NULL;
	}

	if (!info->open) {
		_E("Failed to open usb client open(NULL): %s", id);
		return NULL;
	}

	if (info->open(info, NULL, &common) < 0) {
		_E("Failed to open usb client: %s", id);
		return NULL;
	}

	client = container_of(common, struct usb_client, common);
	if (!client->reconfigure_gadget || !client->enable || !client->disable) {
		_E("Invalid usb client: %s", id);

		if (client->common.info->close)
			client->common.info->close(&client->common);

		return NULL;
	}

	_I("Success lodging usb client: %s", id);
	return client;
}

static int usb_probe(void)
{
	/* Both the gadget and client should be valid. */
	if (gadget_translator || usb_client) {
		if (gadget_translator && usb_client)
			return 0;
		return -EINVAL;
	}

	/* gadget_translator */
	gadget_translator = gadget_translator_probe();
	if (!gadget_translator)
		return -ENODEV;

	/* usb_client */
	usb_client = usb_client_probe(USB_CLIENT_HARDWARE_DEVICE_ID);

	if (!usb_client)
		usb_client = usb_client_probe(USB_CFS_CLIENT_HARDWARE_DEVICE_ID);

	if (!usb_client) {
		if (gadget_translator->common.info)
			gadget_translator->common.info->close(&gadget_translator->common);
		gadget_translator = NULL;

		return -ENODEV;
	}

	return 0;
}

static void usb_release()
{
	if (usb_client) {
		if (usb_client->common.info->close)
			usb_client->common.info->close(&usb_client->common);
		usb_client = NULL;
	}

	if (gadget_translator) {
		if (gadget_translator->common.info->close)
			gadget_translator->common.info->close(
				&gadget_translator->common);
		gadget_translator = NULL;
	}
}

static int usb_config_init(void)
{
	int ret;
	struct usb_gadget_id gadget_id;
	struct usb_gadget *gadget;

	if (!gadget_translator || !usb_client)
		goto no_hal;

	memset(&gadget_id, 0, sizeof(gadget_id));

	gadget_id.function_mask = usb_state_get_selected_mode();
	ret = gadget_translator->id_to_gadget(&gadget_id, &gadget);
	if (ret) {
		_E("Unable to translate id into gadget (%d)", ret);
		return ret;
	}

	usb_client->disable(usb_client);
	ret = usb_client->reconfigure_gadget(usb_client, gadget);
	gadget_translator->cleanup_gadget(gadget);
	if (ret)
		_E("Unable to configure gadget (%d)", ret);

	return ret;
no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static void usb_config_deinit(void)
{
	if (!usb_client)
		goto no_hal;

	usb_client->disable(usb_client);

	return;
no_hal:
	/* TODO. Maybe some default action here? */
	return;
}

static int usb_config_enable(void)
{
	if (!usb_client)
		goto no_hal;

	return usb_client->enable(usb_client);

no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static int usb_config_disable(void)
{
	if (!usb_client)
		goto no_hal;

	usb_client->disable(usb_client);

	return 0;
no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static int usb_change_gadget(unsigned mode)
{
	struct usb_gadget_id gadget_id;
	struct usb_gadget *gadget;
	int ret;

	if (!gadget_translator || !usb_client)
		goto no_hal;

	memset(&gadget_id, 0, sizeof(gadget_id));
	gadget_id.function_mask = mode;

	ret = gadget_translator->id_to_gadget(&gadget_id, &gadget);
	if (ret) {
		_E("Unable to translate id into gadget (%d)", ret);
		return ret;
	}

	usb_client->disable(usb_client);
	ret = usb_client->reconfigure_gadget(usb_client, gadget);
	gadget_translator->cleanup_gadget(gadget);
	if (ret) {
		_E("Unable to configure gadget (%d)", ret);
		goto out;
	}

	_I("USB gadget changed (%d)", mode);

out:
	return ret;
no_hal:
	/* TODO. Maybe some default action here? */
	return -ENODEV;
}

static int usb_enable(unsigned int mode)
{
	int ret;

	if (usb_state_get_connection() != USB_CONNECTED) {
		_E("Failed to enable usb gadget (USB cable is not connected)");
		return -ENOTSUP;
	}

	if (usb_state_get_current_mode() != USB_FUNCTION_NONE)
		usb_state_update_state(USB_CONNECTED, USB_FUNCTION_NONE);

	if (!is_emulator()) {
		ret = usb_config_enable();
		if (ret < 0) {
			_E("Failed to enable usb config (%d)", ret);
			return ret;
		}
	}

	usb_state_update_state(USB_CONNECTED, mode);
	
	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_USB, LCD_OFF, STAY_CUR_STATE, 0);

	return 0;
}

static int usb_disable(void)
{
	int ret;

	if (usb_state_get_current_mode() != USB_FUNCTION_NONE) {
		if (usb_state_get_connection() == USB_CONNECTED)
			usb_state_update_state(USB_CONNECTED, USB_FUNCTION_NONE);
		else
			usb_state_update_state(USB_DISCONNECTED, USB_FUNCTION_NONE);
	}

	if (!is_emulator()) {
		ret = usb_config_disable();
		if (ret != 0)
			_E("Failed to disable usb config (%d)", ret);
	}

	if (disp_plgn.pm_unlock_internal)
		disp_plgn.pm_unlock_internal(INTERNAL_LOCK_USB, LCD_OFF, STAY_CUR_STATE);

	return 0;
}

int usb_change_mode(unsigned mode)
{
	int ret;
	unsigned int curr = usb_state_get_current_mode();

	if (curr != USB_FUNCTION_NONE) {
		ret = usb_disable();
		if (ret < 0) {
			_E("Failed to disable current usb mode");
			return ret;
		}
	}

	if (!is_emulator()) {
		ret = usb_change_gadget(mode);
		if (ret < 0) {
			_E("Failed to change gadget (%d)", ret);
			mode = usb_state_get_selected_mode();
		}
	}
	usb_state_set_selected_mode(mode);

	if (usb_state_get_connection() == USB_CONNECTED) {
		ret = usb_enable(mode);
		if (ret < 0) {
			_E("Failed to enable usb mode (%d)");
			return ret;
		}
	}

	return 0;
}

static int usb_connected(void)
{
	int ret;
	unsigned int mode = usb_state_get_selected_mode();

	usb_state_update_state(USB_CONNECTED, USB_FUNCTION_NONE);

	ret = usb_enable(mode);
	if (ret < 0) {
		_E("Failed to enable usb gadget (gadget:%d, ret:%d)", mode, ret);
		return ret;
	}

	return 0;
}

static int usb_disconnected(void)
{
	usb_state_update_state(USB_DISCONNECTED, USB_FUNCTION_NONE);

	return usb_disable();
}

static int usb_state_changed(int status)
{
	static int old = -1;	/* to update at the first time */
	int ret;

	_I("USB state is changed from (%d) to (%d)", old, status);

	if (old == status)
		return 0;

	switch (status) {
	case USB_CONNECTED:
		_I("USB cable is connected");
		ret = usb_connected();
		break;
	case USB_DISCONNECTED:
		_I("USB cable is disconnected");
		ret = usb_disconnected();
		break;
	default:
		_E("Invalid USB state(%d)", status);
		return -EINVAL;
	}
	if (ret < 0)
		_E("Failed to operate usb connection(%d)", ret);
	else
		old = status;

	return ret;
}

static void usb_init(void *data)
{
	int ret;

	usb_state_retrieve_selected_mode();

	if (!is_emulator()) {
		ret = usb_probe();
		if (ret < 0) {
			_E("USB client cannot be used (%d)", ret);
			return;
		}

		ret = usb_config_init();
		if (ret < 0)
			_E("Failed to initialize usb configuation");
	}

	ret = usb_dbus_init();
	if (ret < 0)
		_E("Failed to init dbus (%d)", ret);

	add_usb_tethering_handler();
	add_usb_debug_handler();

	ret = usb_state_changed(extcon_usb_ops.status);
	if (ret < 0)
		_E("Failed to update usb status(%d)", ret);
}

static void usb_exit(void *data)
{
	remove_usb_debug_handler();
	remove_usb_tethering_handler();
	usb_state_update_state(USB_DISCONNECTED, USB_FUNCTION_NONE);

	if (!is_emulator()) {
		usb_config_deinit();
		usb_release();
	}
}

struct extcon_ops extcon_usb_ops = {
	.name	= EXTCON_CABLE_USB,
	.init	= usb_init,
	.exit	= usb_exit,
	.update = usb_state_changed,
};

EXTCON_OPS_REGISTER(extcon_usb_ops)
